"""
         By Yasuharu Okamoto, 2024.11
"""
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import scipy.optimize as so
from myopt import *             #  極値探索用関数のクラス型
from BlockAverage_v1 import *   #  非計量単調変換用
from copy import copy 

fin_nm = input('Input File Name (*.csv) = ')
with open(fin_nm, 'r') as f:
    raw_data = pd.read_csv(f)
fout_nm = input('Output File Name (*.txt) = ')
fout = open(fout_nm, 'wt')
fout.write('Input Data File = ' + fin_nm + '\n')

obj_names = [v for v in raw_data.keys()[1:]]
print(obj_names)
v_data = raw_data.values
print(v_data)

v_data = v_data[:,1:]

n = len(obj_names)                    #  対象物数
n_ind = len(v_data) // (n + 1)        #  個人数
print('n =', n, '  n_ind =',n_ind)
pers_names = []
for i in range(n_ind):
    pers_names.append(raw_data.values[i * (n + 1)][0])
print('Persons = ', pers_names)

del_ijk = []
for i in range(n_ind):
    del_ijk.append(v_data[1 + (n + 1)*i:(n + 1)*(i+1)])

del_ijk = np.array(del_ijk)
print('Distances =\n', del_ijk)
fout.write(f'Distances =\n{del_ijk}\n')
#
#   初期値設定用の平均距離の配列
#
m_dist = np.mean(del_ijk, axis=0)

D_2 = m_dist**2
J = np.diag([1]*n) - np.ones((n,n))/n
B = J @ (-D_2 / 2) @ J                   #  2重中心化
B = np.array(B, dtype=float)
eig_val, eig_vec = np.linalg.eigh(B)     #  内積行列の固有分解

ck_idx = eig_val > 1.0e-7                #  正の固有値の選択用配列添字

max_r = np.sum(ck_idx)                   #  正の固有値の数

sortidx = np.argsort(-eig_val)           #  固有値の大きいものから降順に並べるための添字配列
eig_val = eig_val[sortidx]
eig_vec = eig_vec[:, sortidx]
#
#　　　正の固有値のスクリープロット
#
plt.title('Eigen Values(>1.0e-7)')
plt.plot(np.arange(1, max_r+1), eig_val[:max_r])
plt.xticks(np.arange(1, max_r+1))
plt.ylabel('Eigen value')
plt.show()

r = int(input(f'Number of dimensions(<={max_r}) = '))
if r > max_r:
    r = max_r
    print(f'Number of dimensions replaced by {r}')
#
#   座標値xと重みw0の初期値
#
x = eig_vec[:,:r] * (eig_val[:r]**0.5)
w0 = np.ones((n_ind, r))

x = x - np.mean(x, axis=0)          #  中心化
s_x = np.max(np.abs(x), axis=0)
x = x / s_x                         #  標準化

x_1 = x         #  初期値
w_1 = w0        #  初期値

d_ijk = copy(del_ijk)
i_nm_t = 0               #  非計量変換の回数

ck_d_hat = [0]*n_ind
ck_del = [0]*n_ind

x_g1 = x_1
w_g1 = w_1
while True:
    i_nm_t += 1
    x_g0 = x_g1
    w_g0 = w_g1
    i_trial = 0      #   非計量変換値に対する極値探索の回数

    d_ijk1 = copy(d_ijk)

    #
    #         非計量変換
    #
    for s in range(n_ind):
        d_delta = np.empty((n * (n - 1)) // 2)
        t_pos = 0
        for i in range(1, n):
            for j in range(i):
                d_delta[t_pos] = del_ijk[s][i][j]
                t_pos += 1

        d_hat = np.empty((n * (n - 1) // 2))
        i_pos = np.empty((n * (n - 1) // 2))
        j_pos = np.empty((n * (n - 1) // 2))
        t_pos = 0
        for i in range(1, n):
            for j in range(i):
                d_hat[t_pos] = d_ijk[s][i][j]
                i_pos[t_pos] = i
                j_pos[t_pos] = j
                t_pos += 1

        d_pair = np.array([(u, v, ipos, jpos) for u, v,ipos, jpos in
                           zip(d_delta, d_hat, i_pos, j_pos)],
                          dtype = ([('delta', float), ('dhat', float),
                                    ('ipos', int), ('jpos', int)]))

        idx = np.argsort(d_pair,order=['delta', 'dhat'])
        dhat = d_pair[idx]['dhat']
        i_pair = d_pair[idx]['ipos']
        j_pair = d_pair[idx]['jpos']

        transf_d = nm_transf(dhat)
        ck_del[s] = d_pair[idx]['delta']
        ck_d_hat[s] = transf_d

        d_ijk1v = np.zeros((n, n))
        

        for p in range((n * (n - 1)) // 2):
            d_ijk1v[i_pair[p]][j_pair[p]] = transf_d[p]
            d_ijk1v[j_pair[p]][i_pair[p]] = transf_d[p]

        d_ijk1v /= np.max(d_ijk1v)
        for j in range(n):
            for k in range(n):
                d_ijk[s][j][k] = d_ijk1v[j][k]


    #
    #        非計量変換値d_ijkに対する極値探索
    #
    while True:
        i_trial +=1
        x_0 = x_1      #  極値探索の初期値
        w_0 = w_1      #  極値探索の初期値
        
        opt_w = for_w(d_ijk**2, n, n_ind, r, x_0)    #  wの極値探索のためのクラス型
        #
        #        wの極値探索
        #
        results = so.minimize(opt_w.f, np.resize(w_0, (n_ind*r)), method='Powell')
        print(results)

        w = (np.reshape(results.x, (n_ind, r))**2)   #  wの設定

        opt_x = for_x(d_ijk**2, n, n_ind, r, w)      #  xの極値探索のためのクラス型
        #
        #        xの極値探索
        #
        results = so.minimize(opt_x.f, np.resize(x_0, (n * r)), method='Powell')
        print(results)

        x = np.reshape(results.x, (n, r))            #  xの設定
        x = x - np.mean(x, axis=0)                   #  xの中心化
        s_x = np.max(np.abs(x), axis=0)              #  標準化のためのスケール値 
        x = x / s_x                                  #  標準化
        w = w * (s_x**2)                             #  xの標準化に合わせた調整
        print('x =\n', x)
        print('w =\n', w)

        x_1 = x     #  極値探索による更新値
        w_1 = w     #  極値探索による更新値

        ck_diff = np.sum((x_1 - x_0)**2) + np.sum((w_1 - w_0)**2)    #  探索前後の変化量
        print(f'ck_diff(at trial {i_trial}) =', ck_diff)

        if ck_diff < (1.0e-7):      #   探索後の変化量が小さいので極値探索を終了する
            break
        if i_trial > 50:            #   十分に探索を行ったとして終了する
            break
    #
    #      更新値に対する距離の計算
    #
    for s in range(n_ind):
        v = x_1 * (w_1[s]**0.5)
        for j in range(n):
            for k in range(n):
                d_ijk1[s][j][k] = (np.sum((v[j] - v[k])**2))**0.5
        v_max = np.max(d_ijk1[s])
        d_ijk1[s] /= v_max                     #   最大値を１に標準化

    x_g1 = x_1
    w_g1 = w_1
    ck_xw = np.sum((x_g1 - x_g0)**2) + np.sum((w_g1 - w_g0)**2)
    print('ck_xw =', ck_xw, '  i_nm_t =', i_nm_t)
    if ck_xw < 1.0e-5:
        fout.write(f'\nck_xw = {ck_xw},   i_nm_t = {i_nm_t}\n')
        break
    ck_diff_nm = np.sum((d_ijk - d_ijk1)**2)
    print('\nck_diff_nm =', ck_diff_nm, '  i_nm_t =', i_nm_t)
    if ck_diff_nm < 1.0e-5:
        fout.write(f'\nck_diff_nm = {ck_diff_nm},  i_nm_t = {i_nm_t}\n')
        break
    if i_nm_t > 20:
        print('i_nm_t =', i_nm_t)
        fout.write(f'\ni_nm_t = {i_nm_t}\n')
        break

    d_ijk = d_ijk1

#
#       Nonmetric trasformationの描画
#
for s in range(n_ind):
    plt.plot(ck_del[s], ck_d_hat[s], label=pers_names[s])
plt.xlabel(r'$\delta$', fontsize=14)
plt.ylabel(r'$\hat{d}$', fontsize=14)
plt.legend()
plt.show()

fout.write('\nCoordinates\n')
fout.write(f'{" ":>10s}')
for i in range(r):
    sv = f'dim-{i+1}'
    fout.write(f'{sv:>10s}')
fout.write('\n')
for i in range(n):
    fout.write(f'{obj_names[i]:>10s}')
    for j in range(r):
        fout.write(f'{x_1[i][j]:>10.5f}')
    fout.write('\n')

fout.write('\n\nWeights\n')
fout.write(f'{" ":>10s}')
for i in range(r):
    sv = f'dim-{i+1}'
    fout.write(f'{sv:>10s}')
fout.write('\n')
for i in range(n_ind):
    fout.write(f'{pers_names[i]:>10s}')
    for j in range(r):
        fout.write(f'{w_1[i][j]:>10.5f}')
    fout.write('\n')

#
#        解の布置の描画
#
xdim = 1
ydim = 2 if r >= 2 else xdim
while True:
    #
    #      項目の座標xの布置
    #
    xdim -= 1
    ydim -= 1
    plt.plot(x_1.T[xdim], x_1.T[ydim], 'o')
    for i in range(n):
        plt.text(x_1.T[xdim][i], x_1.T[ydim][i], obj_names[i])
    plt.xlabel(f'Dim-{xdim+1}')
    plt.ylabel(f'Dim-{ydim+1}')
    plt.title('Coordinates')
    plt.show()
    #
    #       重みの座標wの布置
    #
    plt.plot(w_1.T[xdim], w_1.T[ydim], 'o')
    for i in range(n_ind):
        plt.text(w_1.T[xdim][i], w_1.T[ydim][i], pers_names[i])
    plt.xlabel(f'Dim-{xdim+1}')
    plt.ylabel(f'Dim-{ydim+1}')
    plt.title('Weights')
    plt.show()

    print('Choose dimensions for plot. Set a negative integer when you want to quit.')
    xdim = int(input(f'xdim(<={r}) = '))
    xdim = xdim if xdim <= r else 1
    if xdim < 1:
        break
    ydim = int(input(f'ydim(<={r}) = '))
    ydim = ydim if ydim <= r else r
    if ydim < 1:
        break
#
#       Stressの計算
#
sum_istress = 0.0
for i in range(n_ind):
    dijk = np.empty((n,n))
    xi = x_1 * (w_1[i]**0.5)
    for j in range(n):
        for k in range(n):
            dijk[j][k] = np.sum((xi[j] - xi[k])**2)**0.5
    m_dijk = np.mean(dijk)
    sum_den = np.sum((dijk - m_dijk)**2)
    sum_num = np.sum((dijk - d_ijk[i])**2)
    sum_istress += sum_num / sum_den
stress = (sum_istress / n_ind) ** 0.5

print(f'STRESS = {stress:.5f}')

fout.write(f'\nSTRESS = {stress:.5f}\n')
fout.close()
